#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include <float.h>
#include <fftw3.h>
#include <omp.h>
#include <time.h>

#include "quilting.h"
#include "patch_search.h"
#include "boundary_cut.h"
#include "blending2.h"

#define MAX_DIST 	FLT_MAX

/**
*   This method synthesizes a texture image quilting patches taken from an input sample texture. 
*   @author Lara Raad
*   @param src_im     The input texture sample
*   @param patch_sz   The size of the patches
*   @param overlap_sz The size of the overlap area betwenn patches
*   @param blocks_row Number of patches per row to paste in the synthesized image
*   @param blocks_col Number of patches per column to paste in the synthesized image
*   @param tolerance  The threshold used to select patches from the input image
*   @date 24/05/2012
*/
Image quilting_synthesis(Image src_im, int patch_sz, int overlap_sz, int blocks_row, int blocks_col, float tolerance, Image * position_map, Image * synth_map)
{
    Image out_im;// synthesized image 
    long double *patch_normsH, *patch_normsV, *patch_normsL;
    Corner patch_src;// corner position (x,y) of the chosen patch from src image to paste in the output image
    Corner temp;// temporary corner position (x,y) of the patch under construction
    int step_size;// step size between each patch to paste in the destination image considering the overlap
    int ii, jj, inc_s, inc_o, h_s, w_s, h_o, w_o;
    int fft_size; // size of the r2c ffts for the source image

    //long double min_dist;
    
    int Nc = blocks_col;
	int Nr = blocks_row;
    
    /*************************************************************************/
    /*** INITIALIZATION                                                    ***/
    /*************************************************************************/
    
    //* Initialize ouput image : out_im *//
    out_im  = initialize(src_im, patch_sz, overlap_sz, blocks_row, blocks_col,&patch_src);    
    
    //* Initiialize synthesis map *//
    initialize_image(synth_map, out_im.h, out_im.w, 3);
    
    h_s = src_im.h;
    w_s = src_im.w;
    h_o = out_im.h;
    w_o = out_im.w;
    
    inc_s = h_s*w_s;
    inc_o = h_o*w_o;
    step_size = patch_sz - overlap_sz;
    fft_size = h_s*((int) (w_s/2) + 1);
    
    int h_dict, w_dict;
    h_dict = h_s-patch_sz+1;
    w_dict = w_s-patch_sz+1;
    
    // create the norm of each valid patch in src_im
    patch_normsH = (long double *) malloc(inc_s*sizeof(long double)); // initializing the image containing the norm of
                                                                    // each valid patch in src_im of type horizontal overlap
    patch_normsV = (long double *) malloc(inc_s*sizeof(long double)); // initializing the image containing the norm of
                                                                    // each valid patch in src_im of type vertical overlap
    patch_normsL = (long double *) malloc(inc_s*sizeof(long double)); // initializing the image containing the norm of
                                                                    // each valid patch in src_im of type L-shape overlap
    compute_input_patch_norms(src_im, patch_normsH, patch_normsV, patch_normsL, patch_sz, overlap_sz);
    
    //create the fft of the src_im for each channel
    fftwl_complex *fft_srcR, *fft_srcG, *fft_srcB;
	#pragma omp critical (make_plan)
    fft_srcR = (fftwl_complex*) fftwl_malloc(sizeof(fftwl_complex)*(fft_size));
	#pragma omp critical (make_plan)
    fft_srcG = (fftwl_complex*) fftwl_malloc(sizeof(fftwl_complex)*(fft_size));
	#pragma omp critical (make_plan)
    fft_srcB = (fftwl_complex*) fftwl_malloc(sizeof(fftwl_complex)*(fft_size));
    /* FIXME: remove im_fft */
    im_fft(src_im.img, fft_srcR, h_s, w_s);
    im_fft(src_im.img + inc_s, fft_srcG, h_s, w_s);
    im_fft(src_im.img + 2*inc_s, fft_srcB, h_s, w_s);
    
	Image aux_mask;// Mask of all zeros to initialize the synthesis map
	initialize_image(&aux_mask, patch_sz, patch_sz, 1);

    temp.x = 0;
    temp.y = 0;
    update_synth_map(synth_map, position_map, &patch_src, &temp, patch_sz, aux_mask);

	//* Initializa seeds for parallel loop *//
	unsigned int *seeds;
	size_t num_threads;
	#pragma omp parallel
	{
		num_threads = omp_get_num_threads();
	}
	
	seeds = (unsigned int *)calloc(num_threads, sizeof(unsigned int));
	
	#pragma omp parallel
	{
		#pragma omp for
		for(int kk=0; kk<omp_get_num_threads(); kk++)
		{
			seeds[omp_get_thread_num()] = (omp_get_thread_num()+1)*(unsigned int) time(NULL);
		}
	}
    
    /*************************************************************************/
    /*** MAIN SYNTHESIS LOOP (RUN IN PARALLEL)                             ***/
    /*************************************************************************/
    
    //for(jj=0; jj<(int) w_o - patch_sz + 1 + 2*(Nc-1)*step_size; jj+=step_size)
	for(jj=0; jj<(int) (Nr + 2*(Nc-1) -1)*step_size + 1; jj+=step_size)
    {
        #pragma omp parallel for
		//for(ii=0; ii<(int) h_o - patch_sz + 1; ii+=step_size)
		for(ii=0; ii<(int) (Nc-1)*step_size+1; ii+=step_size)
        {
            //not for the seed patch
            //if ((ii>0 || jj>0) && (jj-2*ii >= 0) && (jj-2*ii < w_o - patch_sz + 1))
			if ((ii>0 || jj>0) && (jj-2*ii >= 0) && (jj-2*ii < w_o - patch_sz + 1))
            {
                Corner temp;
				Corner patch_pos;
				Image boundary_mask;
				
				//* Initialize boundary mask *//
				initialize_image(&boundary_mask, patch_sz, patch_sz, 1);
				
                temp.x = ii;
                temp.y = jj-2*ii;
                
                // Patch search: Randomly pick one of the patch that is close to the minimal distance
                patch_search(src_im, out_im, temp, patch_sz, 
                             patch_normsH, patch_normsV, patch_normsL, 
                             fft_srcR, fft_srcG, fft_srcB, 
                             h_dict, w_dict, &patch_pos, tolerance, &seeds[omp_get_thread_num()]);

				// compute boundary mask
				compute_boundary_mask(src_im, out_im, &boundary_mask, 
									  overlap_sz, patch_sz, &temp, patch_pos, 
									  &seeds[omp_get_thread_num()]);
									
				// Updating synthesis map
                update_synth_map(synth_map, position_map, &patch_pos, &temp, patch_sz, boundary_mask);					
				
                // blending new patch into out_im
                blend_patch(boundary_mask, &out_im, src_im, patch_pos, temp, patch_sz, overlap_sz);

				free(boundary_mask.img);
                
            }
            
        }        
    }

    
    //free memory
    free(patch_normsH);
    free(patch_normsV);
    free(patch_normsL);
	free(seeds);
	free(aux_mask.img);
	#pragma omp critical (make_plan)
    fftwl_free(fft_srcR);
	#pragma omp critical (make_plan)
    fftwl_free(fft_srcG);
	#pragma omp critical (make_plan)
    fftwl_free(fft_srcB);
    
    return out_im;
}