#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <string.h>
#include <fftw3.h>

#include "quilting_lib.h"
#include "quilting.h"

/**
*   This method initializes the output image taking a random patch from the input
*   texture and placing it on the left-top corner of the output image.
*   @author Lara Raad
*   @param src_im      The input texture sample
*   @param patch_sz    The size of the patches
*   @param overlap_sz  The size of the overlap area betwenn patches
*   @param blocks_row  Number of patches per row to synthesize in the output image
*   @param blocks_col  Number of patches per column to synthesize in the output image
*   @param patch_src   The top left corner position of the patch copied from the source
*   @date 24/05/2012
*/
Image initialize(Image src_img, int patch_sz, int overlap_sz, int blocks_row, int blocks_col, Corner *patch_src){
    Image img;
    int i, j, k, rows, cols, channels, pos, pos_s, inc, inc_s;
    rows = blocks_row*patch_sz - (blocks_row - 1)*overlap_sz;
    cols = blocks_col*patch_sz - (blocks_col - 1)*overlap_sz;
    channels = src_img.c;

    initialize_image(&img, rows, cols, channels);
      
    // Paste random block from input texture in left top corner
    // of the output image.
     
    // Random positions in [0,h-patch_sz)x[O,w-patch_sz)
	int x_rand = rand()%(src_img.h-patch_sz+1);
	assert(x_rand<(int)(src_img.h-patch_sz+1));
	int y_rand = rand()%(src_img.w-patch_sz+1);
	assert(y_rand<(int)(src_img.w-patch_sz+1));
	
	patch_src->x = x_rand;
    patch_src->y = y_rand;

    inc = img.h*img.w;
    inc_s = src_img.h*src_img.w;
    
    //paste the first block into the output image
    for(i=0; i < patch_sz; i++)
    {
        for(j=0; j < patch_sz; j++)
        {
            pos = i*img.w + j; 
            pos_s = (x_rand + i)*src_img.w +(y_rand +j);
			assert(pos_s < (int) (src_img.w*src_img.h));
			for(k=0; k<channels; k++)
			{
				img.img[pos + k*inc] = src_img.img[pos_s + k*inc_s];
			}
        }
    }
            
    return img;
}

/**
*   This method initializes images of type float. It sets the numeber of rows, columns and channels 
*   and sets the pixels values to zero.
*   @author Lara Raad
*   @param im       The image to initialize
*   @param rows     Number of rows of the image
*   @param cols     Number of columns of the image
*   @param channels Number of channels of the image
*   @date 24/05/2012
*/
void initialize_image(Image *im, int rows, int cols, int channels)
{
    int size;
    im->h = rows;
    im->w = cols;
    im->c = channels;
    size = rows*cols*channels;
    im->img = (float *) calloc(size, sizeof(float));
}

/**
*   This method initializes images of type double. It sets the numeber of rows, columns and channels 
*   and sets the pixels values to zero.
*   @author Lara Raad
*   @param im       The image to initialize
*   @param rows     Number of rows of the image
*   @param cols     Number of columns of the image
*   @param channels Number of channels of the image
*   @date 24/05/2012
*/
void initialize_image_D(Image_D *im, int rows, int cols, int channels)
{
    int size;
    im->h = rows;
    im->w = cols;
    im->c = channels;
    size = rows*cols*channels;
    im->img = (double *) calloc(size, sizeof(double));
}

/**
*   This method initializes images of type long double. It sets the numeber of rows, columns and channels 
*   and sets the pixels values to zero.
*   @author Lara Raad
*   @param im       The image to initialize
*   @param rows     Number of rows of the image
*   @param cols     Number of columns of the image
*   @param channels Number of channels of the image
*   @date 24/05/2012
*/
void initialize_image_LD(Image_LD *im, int rows, int cols, int channels)
{
    int size;
    im->h = rows;
    im->w = cols;
    im->c = channels;
    size = rows*cols*channels;
    im->img = (long double *) calloc(size, sizeof(long double));
}

/**
*   This method initializes images of type unsigned long. It sets the numeber of rows, columns and channels 
*   and sets the pixels values to zero.
*   @author Lara Raad
*   @param im       The image to initialize
*   @param rows     Number of rows of the image
*   @param cols     Number of columns of the image
*   @param channels Number of channels of the image
*   @date 24/05/2012
*/
void initialize_image_UI(Image_UI *im, int rows, int cols, int channels)
{
    int size;
    im->h = rows;
    im->w = cols;
    im->c = channels;
    size = rows*cols*channels;
    im->img = (unsigned long *) calloc(size, sizeof(unsigned long));
}
/**
*   This method creates a position map that associates a different
*	color from a continuous colormap to every pixel of the source 
*	image
*   @author Lara Raad
*   @param position_map     Position map image
*   @date 09/05/2013
*/
void create_position_map(Image position_map)
{
    size_t count = 0;
    size_t inc = position_map.h*position_map.w;
    size_t i,j;
    
    for(i=0;i<position_map.h;i++)
    {
        for(j=0;j<position_map.w;j++)
        {
            position_map.img[count] = (i*255)/position_map.h;//Red channel
            position_map.img[count+inc] = (j*255)/position_map.w;//Green channel
            position_map.img[count+2*inc] = (i*j*255)/(position_map.w*position_map.h);//Blue channel
            count++;
        }
    }
}

/**
*   This method crops an image to be of size MxNxC
*   @author Lara Raad
*   @param im       Image to crop
*   @param M,N,C    Size of the cropped image
*   @param im_crop  The cropped image
*   @date 09/08/2013
*/

int crop_image(Image *im_crop, Image *im, int M, int N, int C)
{
    if( (M > (int) im->h) || (N > (int) im->w) )
    {
        return 1;
    }
    
    int i,j,k;
    initialize_image(im_crop, M, N, C);
    for(k=0;k<C;k++)
    {
        for(i=0;i<M;i++)
        {
            for(j=0;j<N;j++)
            {
                im_crop->img[i*N+j + k*M*N] = im->img[i*im->w+j + k*im->h*im->w];
            }
        }
    }        
    return 0;
    
}

/**
*   This method updates the synthesis map for each patch added
*	in the synthesized image
*   @author Lara Raad
*   @param synth_map			Synthesis map
*   @param position_map			Color map of the texture sample
*   @param leftTopCorner_src	The position of the patch to add taken in the source image
*   @param leftTopCorner_output	The position in the ouput image where we add a block
*   @param pathc_sz				The patch size
*   @date 26/02/2014
*/
int update_synth_map(Image * synth_map, Image * position_map, Corner * leftTopCorner_src, Corner * leftTopCorner_output, int patch_sz, Image boundary_mask)
{
	int kk, ll, pos_posMap, pos_synthMap, x_s, y_s, x_o, y_o, inc_o, inc_s;
	x_s = leftTopCorner_src->x;
	y_s = leftTopCorner_src->y;
	assert( (x_s < (int)((synth_map->w-patch_sz+1)*(synth_map->h-patch_sz+1))) && (y_s < (int)((synth_map->w-patch_sz+1)*(synth_map->h-patch_sz+1))) );
	x_o = leftTopCorner_output->x;
	y_o = leftTopCorner_output->y;
	inc_s = position_map->w*position_map->h;
	inc_o = synth_map->w*synth_map->h;
	
	for (kk=0; kk<patch_sz; kk++)
    {
        for (ll=0; ll<patch_sz; ll++)
        {
            pos_posMap = (x_s + kk)*position_map->w + (y_s + ll);
            pos_synthMap = (x_o + kk)*synth_map->w + (y_o + ll);
			assert(pos_posMap < (int)(position_map->h*position_map->w));
			assert(pos_synthMap < (int)(synth_map->h*synth_map->w));
            synth_map->img[pos_synthMap] = boundary_mask.img[kk*boundary_mask.w+ll]*synth_map->img[pos_synthMap] 
											+ (1-boundary_mask.img[kk*boundary_mask.w+ll])*position_map->img[pos_posMap];
            synth_map->img[pos_synthMap + inc_o] = boundary_mask.img[kk*boundary_mask.w+ll]*synth_map->img[pos_synthMap + inc_o] 
													+ (1-boundary_mask.img[kk*boundary_mask.w+ll])*position_map->img[pos_posMap + inc_s];
            synth_map->img[pos_synthMap + 2*inc_o] = boundary_mask.img[kk*boundary_mask.w+ll]*synth_map->img[pos_synthMap + 2*inc_o]
														+ (1-boundary_mask.img[kk*boundary_mask.w+ll])*position_map->img[pos_posMap + 2*inc_s];
        }
    }

	return 0;
}

/**
*   This method removes the extension of a filename
*   @author Lara Raad
*   @param filiename			The file name to which the extension is removed
*   @date 25/08/2014
*/
char* remove_ext_filename(char *filename)
{
	//get the string after the last '/' (when exists) and remove extension .png
	char *file;
	if ((file = strrchr( filename,  '/'))==NULL)
		file = filename;
	else
		file = file +1;
	int i = strlen(file);
	while( (file[i-1] != '.') && (i>0))
	{
		i--;	
	}
	int posDot = i-1;
	char *fileNoExt = (char *) calloc(posDot, sizeof(char));
	strncpy(fileNoExt,file,posDot);
	return fileNoExt;	
}

/**
*   This method computes the 2-D FFT of an input image
*   @author Lara Raad
*   @param img			Input image to apply the 2-D FFT
*   @param fft_im		2-D FFT of the input image img
*   @param h,w			Dimensions of img (height and width)
*   @date 26/02/2014
*/
int im_fft(float *img, fftwl_complex *fft_im, int h, int w)
{
	fftwl_plan p_im;
	Image_LD src_ld;// copy of src_im in a long double version
	int pos_im;
        
	initialize_image_LD(&src_ld, h, w, 1);
	
	for(int i=0; i<(int)h; i++)
	{
		for(int j=0; j<(int)w; j++)
		{
			pos_im = i*w + j;
			src_ld.img[pos_im] = (long double)(img[pos_im]);
		}
	}
	
	// create plan
	#pragma omp critical (make_plan)
	p_im = fftwl_plan_dft_r2c_2d(h,w,src_ld.img,fft_im,FFTW_ESTIMATE);
	// execute fft
	fftwl_execute(p_im);
	
	free(src_ld.img);
	#pragma omp critical (make_plan)
	fftwl_destroy_plan(p_im);
	
	return 0;
}